#include "plugin.h"
#include "plugapi.h"
#include <dlfcn.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <pthread.h>

/**
 * TODO
 * ====
 *
 * - Expose the error returned by dlerror() in some way. It could be stored in
 *   the mhandle struct, for example.
 */

/**
 * struct mhandle
 *
 *   The definition of the mhandle struct. This struct is returned as a handle
 *   from plugin_open().
 */
struct mhandle
{
	void *handle;     /* handle returned by dlopen() */

	char *filename;   /* the plugin filename */

	struct motion_plugin *
	pluginfo;         /* the plugin info struct present in the plugin */
};

/**
 * thread_lock
 *
 *   This is a mutex that is used to make the plugin-handling functions 
 *   thread-safe. The dlerror() function is not reentrant, so we need to 
 *   protect all dl* functions from concurrent access.
 */
pthread_mutex_t thread_lock = PTHREAD_MUTEX_INITIALIZER;

/* plugin_data - returns the plugin info struct */
struct motion_plugin *plugin_data(struct mhandle *handle)
{
	return handle->pluginfo;
}

/**
 * plugin_resolve_nolock
 *
 *   Resolves a symbol in the plugin, but without locking the mutex. This
 *   function is called from the other plugin functions, as they need to
 *   resolve symbols but handle locking themselves. We could also use an
 *   error-checking mutex and check for a deadlock situation, but this
 *   solution works.
 *
 * Parameters:
 *
 *   handle  - the plugin handle
 *   symbol  - the name of the symbol to resolve
 *   address - where to store the address of the resolved symbol
 *
 * Returns: 0 on failure, 1 on success
 */
int plugin_resolve_nolock(struct mhandle *handle, const char *symbol, 
                          void **address)
{
	void *addr;
	char *error;

	/* dlsym() may return a NULL address, which may be ok. Thus, to check for
	 * error, we have to clear the error with dlerror(), and then see if 
	 * dlerror() returns non-NULL after the call to dlsym().
	 */
	dlerror(); /* clear error */
	addr = dlsym(handle->handle, symbol);
	if ((error = dlerror()) != NULL) {
		fprintf(stderr, "Symbol lookup failed: %s\n", error);
		*address = NULL;
		return 0;
	}
	
	*address = addr;
	return 1;
}

/* plugin_load - load/open the specified plugin */
struct mhandle *plugin_load(const char *filename, int lazy)
{
	void *handle;
	struct mhandle *ret = NULL;
	struct motion_plugin *modinfo;
	plugin_load_function fnload;

	pthread_mutex_lock(&thread_lock);
	
	/* Try to open the library that contains the plugin. */
	handle = dlopen(filename, lazy ? RTLD_LAZY : RTLD_NOW);
	if (!handle) {
		fprintf(stderr, "Failed to load the plugin '%s'.\n", filename);
		/* I know Kenneth hates gotos, but it's quite handy for error-checking
		 * in this function...
		 */
		goto err;
	}

	/* Allocate the plugin handle struct. */
	ret = malloc(sizeof(struct mhandle)); //mymalloc
	memset(ret, 0, sizeof(struct mhandle)); /* reset */
	ret->handle = handle;

	/* Get the plugin info, which is mandatory. */
	if (!plugin_resolve_nolock(ret, M_INFO_SYMBOL, (void **)&modinfo)) {
		fprintf(stderr, "No plugin info found in plugin '%s'.\n", filename);
		goto err;
	}

	ret->pluginfo = modinfo;
	ret->filename = strdup(filename);
	
	/* Call the load function if it exists in the plugin (it's ok if it 
	 * isn't there).
	 */
	if (plugin_resolve_nolock(ret, M_LOAD_FUNC, (void **)&fnload)) {
		if (!(*fnload)()) {
			/* Failure return from the load function. */
			fprintf(stderr, "The plugin load function returned with failure.\n");
			goto err;
		}
	}

	pthread_mutex_unlock(&thread_lock);
	return ret;

err:
	if (ret) {
		if (ret->handle) {
			dlclose(ret->handle);
		}
		free(ret);
	}
	pthread_mutex_unlock(&thread_lock);
	return NULL;
}

/* plugin_close - closes/unloads the specified plugin */
void plugin_close(struct mhandle *handle)
{
	plugin_unload_function fnunload;

	pthread_mutex_lock(&thread_lock);

	/* Call the unload function if it exists. */
	if (plugin_resolve_nolock(handle, M_UNLOAD_FUNC, (void **)&fnunload)) {
		(*fnunload)();
	}
	
	/* Free stuff and close the native dl handle... */
	free(handle->filename);
	dlclose(handle->handle);
	free(handle);

	pthread_mutex_unlock(&thread_lock);
}

/* plugin_resolve - resolves a symbol (wrapper around plugin_resolve_nolock) */
int plugin_resolve(struct mhandle *handle, const char *symbol, void **address)
{
	int ret;
	
	pthread_mutex_lock(&thread_lock);
	
	ret = plugin_resolve_nolock(handle, symbol, address);

	pthread_mutex_unlock(&thread_lock);

	return ret;
}

